# coding=utf-8
'''
pymaze
Copyright (C) 2012-2014 Moses Palmér

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
'''

import math

from . import BaseWall, BaseMaze

class TriWall(BaseMaze.Wall):
    # The angles for each corner; this is a list of the tuple (angle, alt_angle)
    # where alt_angle is used for the room at x, y when x + y is odd
    _ANGLES = []

    # The directions for each wall; this is a list of the tuple (dir, alt_dir)
    # where alt_dir is used for the room at x, y when x + y is odd
    _DIRECTIONS = []

    # The names of the walls
    NAMES = []

    # The list of walls; this is the list [0, 1, 2... N]
    WALLS = []

    # The horizontal scale factor when converting maze coordinates to physical
    # coordinates
    HORIZONTAL_MULTIPLICATOR = math.cos(math.pi / 2 + 2 * 2 * math.pi / 3)

    # The vertical scale factor when converting maze coordinates to physical
    # coordinates
    VERTICAL_MULTIPLICATOR = 2 + math.sin(math.pi / 2 + 2 * math.pi / 3)

    # The vertical offset added or removed when converting maze coordinates to
    # physical coordinates; it is added if x + y is odd
    OFFSET = 0.5 * (1 + math.sin(math.pi / 2 + 2 * math.pi / 3))

    __slots__ = tuple()

    start_angle = math.pi / 2 + 2 * math.pi / 3
    data = (
        ('DIAGONAL_1', (-1, 0), (1, 0)),
        ('DIAGONAL_2', (1, 0), (-1, 0)),
        ('HORIZONTAL', (0, -1), (0, 1)))
    for i, (name, dir1, dir2) in enumerate(data):
        locals()[name] = i

        next_angle = _ANGLES[-1][0] - 2 * math.pi / len(data) \
            if _ANGLES else start_angle
        while next_angle < 0.0:
            next_angle += 2 * math.pi
        alt_angle = next_angle - math.pi
        while alt_angle < 0.0:
            alt_angle += 2 * math.pi
        _ANGLES.append((next_angle, alt_angle))

        _DIRECTIONS.append((dir1, dir2))

        NAMES.append(name.lower())
        WALLS.append(i)

    @classmethod
    def from_direction(self, room_pos, direction):
        """
        @see Maze.Wall.from_direction
        """
        alt = (room_pos[0] + room_pos[1]) % 2
        for i, dirs in enumerate(self._DIRECTIONS):
            if direction == dirs[alt]:
                return self(room_pos, i)

        raise ValueError('Invalid direction for %s: %s' % (
            str(room_pos), str(direction)))

    @classmethod
    def from_corner(self, room_pos, wall_index):
        """
        @see Maze.Wall.from_corner
        """
        start_wall = wall = self(room_pos, wall_index)

        while True:
            yield wall

            back = wall.back
            next_room_pos = back.room_pos
            next_wall = (int(back) + 1) % len(self.WALLS)
            next = self(next_room_pos, next_wall)
            if next == start_wall:
                break

            wall = next

    def _get_opposite_index(self):
        """
        There is no opposite wall in a triangular room.

        @raise NotImplementedError
        """
        raise NotImplementedError()

    def _get_back_index(self):
        """
        @see Maze.Wall._get_back_index
        """
        return self.wall

    def _get_direction(self):
        """
        @see Maze.Wall._get_direction
        """
        alt = (self.room_pos[0] + self.room_pos[1]) % 2
        return self._DIRECTIONS[self.wall][alt]

    def _get_span(self):
        """
        @see Maze.Wall._get_span
        """
        alt = (self.room_pos[0] + self.room_pos[1]) % 2
        start = self._ANGLES[self.wall][alt]
        end = self._ANGLES[(self.wall + 1) % len(self._ANGLES)][alt]

        return (start, end)

class TriMaze(BaseMaze):
    """A maze with triangular rooms.

    The rooms alternate between having a wall facing downwards and a corner
    facing downwards. The room at x, y will have a corner facing downwards if
    x + y is odd.

    Walls in a triangular maze do not have an opposite wall.
    """
    Wall = TriWall

    def get_center(self, room_pos):
        alt = (room_pos[0] + room_pos[1]) % 2
        sign = 1 if alt == 1 else -1
        return (
            (room_pos[0] + 0.5) * self.Wall.HORIZONTAL_MULTIPLICATOR,
            (room_pos[1] + 0.5) * self.Wall.VERTICAL_MULTIPLICATOR \
                + sign * self.Wall.OFFSET)



# coding=utf-8
'''
pymaze
Copyright (C) 2012-2014 Moses Palmér

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
'''

import sys

def print_maze(maze, solution, wall_char = '@', path_char = '.',
        floor_char = ' ', room_size = (5, 4)):
    """
    Prints a maze and its solution.

    This function works only for mazes with square rooms.

    @param maze
        The maze to print. If this is not a maze with square rooms, False is
        immediately returned.
    @param solution
        The solution, as a list generated by list(maze[start:end]).
    @param wall_char, path_char, floor_char
        The characters to use to print the walls, the solution path and the
        floor. The lengths of each of these string must be equal.
    @param room_size
        The size, expressed as (width, height), of a room in characters.
    @return True if the maze could be printed and False otherwise
    """
    if len(maze.Wall.WALLS) != 4:
        print('This maze cannot be printed as it is not square')
        return False

    def output(s):
        sys.stdout.write(s)

    # Iterate over all rows and make sure to start with the last one to maintain
    # the orientation of the maze
    for y in reversed(range(maze.height)):
        # Print one line for each height unit of a room
        for line in range(room_size[1]):
            # Calculate the vertical offset of the neighbouring room
            dy = 1 if not line else -1 if line == room_size[1] - 1 else 0
            for x in range(maze.width):
                # Print one line of the current room for every room in the
                # current row
                if line == 0 or line == room_size[1] - 1:
                    # This is the first or the last line
                    output(''.join((
                        # The left wall
                        wall_char,

                        # The opening or top/bottom wall
                        (path_char if \
                            int(maze.Wall.from_direction((x, y), (0, dy))) \
                                in maze[x, y] \
                            and (x, y) in solution \
                            and (x, y + dy) in solution
                        else floor_char \
                            if int(maze.Wall.from_direction((x, y), (0, dy))) \
                                in maze[x, y]
                        else wall_char) * (room_size[0] - 2),

                        # The right wall
                        wall_char)))
                else:
                    # This is a line in the middle of a room
                    output(''.join((
                        # The left opening or wall
                        path_char if int(maze.Wall.from_direction(
                                (x, y), (-1, 0))) in maze[x, y] \
                            and (x, y) in solution \
                            and (x - 1, y) in solution
                        else floor_char if \
                            int(maze.Wall.from_direction((x, y), (-1, 0))) \
                                in maze[x, y]
                        else wall_char,

                        # The center
                        (path_char if (x, y) in solution \
                        else floor_char) * (room_size[0] - 2),

                        # The right opening or wall
                        path_char if int(maze.Wall.from_direction(
                                (x, y), (1, 0))) in maze[x, y] \
                            and (x, y) in solution \
                            and (x + 1, y) in solution
                        else floor_char if \
                            int(maze.Wall.from_direction((x, y), (1, 0))) \
                                in maze[x, y]
                        else wall_char)))
            output('\n')

    return True




# coding=utf-8
'''
pymaze
Copyright (C) 2012-2014 Moses Palmér

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
'''

import math
import os
import random
import re
import sys

try:
    import cairocffi as cairo
except ImportError:
    try:
        import cairo
    except ImportError:
        print('This program requires cairo')
        sys.exit(1)

from maze.quad import Maze
from maze.tri import TriMaze
from maze.hex import HexMaze
from maze.randomized_prim import initialize

from .terminal import print_maze
from .image import make_image

def main():
    import argparse

    parser = argparse.ArgumentParser(
        description = 'A tool to generate mazes')

    def maze_size(s):
        result = int(s)
        if result < 1:
            raise argparse.ArgumentTypeError(
                'The maze size must be greater than 0')
        else:
            return result
    parser.add_argument('--maze-size', type = maze_size, nargs = 2,
        metavar = ('WIDTH', 'HEIGHT'),
        default = (15, 10),
        help = 'The size of the maze.')

    maze_classes = dict((len(mc.Wall.WALLS), mc) for mc in (
        Maze, TriMaze, HexMaze))
    parser.add_argument('--walls', type = int,
        choices = maze_classes.keys(),
        default = 4,
        help = 'The number of walls for every room.')

    def char(s):
        if len(s) != 1:
            raise argparse.ArgumentTypeError('%s is not a valid character' % s)
        else:
            return s
    parser.add_argument('--print-wall-char', type = char,
        default = '@',
        help = 'The character used for walls when printing the maze.')
    parser.add_argument('--print-path-char', type = char,
        default = '.',
        help = 'The character used for the path when printing the maze.')
    parser.add_argument('--print-floor-char', type = char,
        default = ' ',
        help = 'The character used for the floor when printing the maze.')

    def print_room_size(s):
        result = int(s)
        if result < 3:
            raise argparse.ArgumentTypeError(
                'The room size must be greater than or equal to 3')
        else:
            return result
    parser.add_argument('--print-room-size', type = print_room_size,
        nargs = 2,
        default = (5, 4),
        help = 'The size of each room in characters when printing the maze.')

    def image_room_size(s):
        result = int(s)
        if result < 1:
            raise argparse.ArgumentTypeError(
                'The maze room size in the image must be greater than 0')
        else:
            return int(0.5 * math.sqrt(2.0) * result)
    parser.add_argument('--image-room-size', type = image_room_size, nargs = 2,
        metavar = ('WIDTH', 'HEIGHT'),
        default = (30, 30),
        help = 'The size of the rooms in the maze image.')

    surface_types = {
        'pdf': (
            lambda f, w, h: cairo.PDFSurface(f, w, h),
            lambda f, surface: None),
        'png': (
            lambda f, w, h: cairo.ImageSurface(cairo.FORMAT_ARGB32, w, h),
            lambda f, surface: surface.write_to_png(f)),
        'ps': (
            lambda f, w, h: cairo.PSSurface(f, w, h),
            lambda f, surface: None),
        'svg': (
            lambda f, w, h: cairo.SVGSurface(f, w, h),
            lambda f, surface: None)}
    def surface(s):
        try:
            ext = s.rsplit(os.path.extsep, 1)[1]
            return (
                lambda w, h: surface_types[ext][0](s, w, h),
                lambda surface: surface_types[ext][1](s, surface))
        except KeyError as e:
            argparse.ArgumentTypeError(
                '"%s" is not a valid file extension' % e.args[0])
        except IndexError:
            argparse.ArgumentTypeError(
                'The image file must have a valid extension')
    class default:
        def write(*args, **kwargs):
            pass
    parser.add_argument('--image-output', type = surface,
        metavar = 'FILENAME',
        required = True,
        help = ('The name of the image file to create. Valid types are %s.') % (
            ', '.join(surface_types.keys())))

    def color(allow_rgba):
        def rgb(s):
            result = None
            m = re.match(r'''(?x)
                \#(?P<red>[0-9A-Fa-f]{2})
                 (?P<green>[0-9A-Fa-f]{2})
                 (?P<blue>[0-9A-Fa-f]{2})''', s)
            if not m is None:
                result = tuple(float(int(d, 16)) / 255 for d in m.groups())
            else:
                m = re.match(r'''(?x)
                    rgb\(
                        \s*(?P<red>\d{1,3}%?)\s*,
                        \s*(?P<green>\d{1,3}%?)\s*,
                        \s*(?P<blue>\d{1,3}%?)\s*\)''', s)
                if not m is None:
                    result = tuple(float(d) / 255 if d[-1].isdigit()
                        else float(d[:-1]) / 100 for d in m.groups())
            if result is None or any(r < 0 or r > 1.0 for r in result):
                raise argparse.ArgumentTypeError(
                    '"%s" is not a valid colour.' % s)
            return result + (1.0,)
        def rgba(s):
            result = None
            m = re.match(r'''(?x)
                \#(?P<alpha>[0-9A-Fa-f]{2})
                 (?P<red>[0-9A-Fa-f]{2})
                 (?P<green>[0-9A-Fa-f]{2})
                 (?P<blue>[0-9A-Fa-f]{2})''', s)
            if not m is None:
                result = tuple(float(int(d, 16)) / 255
                    for d in m.groups()[1:] + (m.group(1),))
            else:
                m = re.match(r'''(?x)
                    rgba\(
                        \s*(?P<red>\d{1,3}%?)\s*,
                        \s*(?P<green>\d{1,3}%?)\s*,
                        \s*(?P<blue>\d{1,3}%?)\s*\,
                        \s*(?P<alpha>\d{1,3}%?)\s*\)''', s)
                if not m is None:
                    result = tuple(float(d) / 255 if d[-1].isdigit()
                        else float(d[:-1]) / 100 for d in m.groups())
            if result is None:
                return rgb(s)
            elif any(r < 0 or r > 1.0 for r in result):
                raise argparse.ArgumentTypeError(
                    '"%s" is not a valid colour.' % s)
            return result
        if allow_rgba:
            return rgba
        else:
            return rgb
    parser.add_argument('--image-background-color', type = color(True),
        metavar = 'COLOUR',
        default = (0.0, 0.0, 0.0),
        help = 'The background colour of the image. This must be specified as '
            'an HTML colour on the form #RRGGBB or rgb(r, g, b), or #AARRGGBB '
            'or rgba(r, g, b, a).')
    parser.add_argument('--image-wall-color', type = color(False),
        metavar = 'COLOUR',
        default = (1.0, 1.0, 1.0),
        help = 'The colour of the wall in the image. This must be specified as '
            'an HTML colour on the form #RRGGBB or rgb(r, g, b).')
    parser.add_argument('--image-path-color', type = color(True),
        metavar = 'COLOUR',
        default = (0.8, 0.4, 0.2),
        help = 'The colour of the path in the image. This must be specified as '
            'an HTML colour on the form #RRGGBB or rgb(r, g, b), or #AARRGGBB '
            'or rgba(r, g, b, a).')

    def line_width(s):
        result = int(s)
        if result < 2:
            raise argparse.ArgumentTypeError(
                'Line widths must be greater than 1')
        elif result & 1:
            raise argparse.ArgumentTypeError(
                'Line widths must be an even number')
        else:
            return result
    parser.add_argument('--image-wall-width', type = line_width,
        metavar = 'WIDTH',
        default = 2,
        help = 'The width of the maze wall lines.')
    parser.add_argument('--image-path-width', type = line_width,
        metavar = 'WIDTH',
        default = 2,
        help = 'The width of the maze path lines.')

    parser.add_argument('--image-path-smooth',
        action = 'store_true',
        default = False,
        help = 'Whether the path should be painted as a smooth curve instead '
            'of a sharp line.')

    namespace = parser.parse_args()

    # Create and initialise the maze
    maze = maze_classes[namespace.walls](*namespace.maze_size)
    initialize(maze, lambda max: random.randint(0, max - 1))
    solution = list(maze.walk_path((0, 0), (maze.width - 1, maze.height - 1)))

    print_maze(maze, solution, **dict(
            (name.split('_', 1)[1], value)
                for name, value in vars(namespace).items()
                if name.startswith('print_')))
    make_image(maze, solution, **dict(
            (name.split('_', 1)[1], value)
                for name, value in vars(namespace).items()
                if name.startswith('image_')))




# coding=utf-8
'''
pymaze
Copyright (C) 2012-2014 Moses Palmér

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
'''

import math

from . import BaseWall, BaseMaze

class HexWall(BaseMaze.Wall):
    # The angles for each corner; this is a list of the tuple (angle, alt_angle)
    # where alt_angle is used for the room at x, y when x + y is odd
    _ANGLES = []

    # The directions for each wall; this is a list of the tuple (dir, alt_dir)
    # where alt_dir is used for the room at x, y when x + y is odd
    _DIRECTIONS = []

    # The names of the walls
    NAMES = []

    # The list of walls; this is the list [0, 1, 2... N]
    WALLS = []

    # The horizontal scale factor when converting maze coordinates to physical
    # coordinates
    HORIZONTAL_MULTIPLICATOR = 2.0 * math.cos(math.pi / 6)

    # The vertical scale factor when converting maze coordinates to physical
    # coordinates
    VERTICAL_MULTIPLICATOR = 2.0 - math.sin(math.pi / 6)

    __slots__ = tuple()

    start_angle = math.pi / 2 + (2 * 2 * math.pi) / 6
    data = (
        ('LEFT', (-1, 0), None),
        ('UP_LEFT', (-1, 1), (0, 1)),
        ('UP_RIGHT', (0, 1), (1, 1)),
        ('RIGHT', (1, 0), None),
        ('DOWN_RIGHT', (0, -1), (1, -1)),
        ('DOWN_LEFT', (-1, -1), (0, -1)))
    for i, (name, dir1, dir2) in enumerate(data):
        locals()[name] = i

        next_angle = _ANGLES[-1] - 2 * math.pi / len(data) \
            if _ANGLES else start_angle
        while next_angle < 0.0:
            next_angle += 2 * math.pi
        _ANGLES.append(next_angle)

        _DIRECTIONS.append((dir1, dir2))

        NAMES.append(name.lower())
        WALLS.append(i)

    @classmethod
    def from_direction(self, room_pos, direction):
        """
        @see Maze.Wall.from_direction
        """
        use_alt = room_pos[1] % 2 == 1
        for i, (dir1, dir2) in enumerate(self._DIRECTIONS):
            if direction == (dir1 if not use_alt or not dir2 else dir2):
                return self(room_pos, i)

        raise ValueError('Invalid direction for %s: %s' % (
            str(room_pos), str(direction)))

    def _get_direction(self):
        """
        @see Maze.Wall._get_direction
        """
        if (self.room_pos[1] % 2 == 1) and self._DIRECTIONS[self.wall][1]:
            return self._DIRECTIONS[self.wall][1]
        else:
            return self._DIRECTIONS[self.wall][0]

class HexMaze(BaseMaze):
    """A maze with hexagonal rooms.

    The rooms are laid out in rows, where every odd row is moved a half room in
    the positive horizontal direction; the appearance of the maze is that of a
    honey comb.
    """
    Wall = HexWall

    def get_center(self, room_pos):
        return (
            (room_pos[0] + (1.0 if room_pos[1] % 2 == 1 else 0.5))
                * self.Wall.HORIZONTAL_MULTIPLICATOR,
            (room_pos[1] + 0.5) * self.Wall.VERTICAL_MULTIPLICATOR)



# coding=utf-8
'''
pymaze
Copyright (C) 2012-2014 Moses Palmér

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
'''

import bisect
import math
import sys

from . import _info


class BaseWall(object):
    """A reference to the wall of a room.

    A wall has an index, a direction, a span and a position.
      * The index is its position in the list of walls. This corresponds to its
        symbolic name.
      * The direction is a direction vector for the wall; up and right are
        positive directions. The direction is the movement vector in the maze
        room matrix and not necessarily an actual physical direction.
      * The span is the physical start and end angle of the wall.
      * The position is the coordinates of the room to which the wall belongs in
        the maze room matrix.

    Furthermore, walls know their back and opposite.
      * The back is the same wall in the room on the other side of a wall. A
        wall will compare equally with its back.
      * The opposite is the wall opposing a wall in the same room. Depending on
        the layout of a room, there may be no opposite wall. Triangular mazes do
        not have opposite walls.

    Walls have factory methods.
      * from_direction will create a wall when the room and the direction is
        known. The direction must be the direction vector of the room.
      * from_room_pos will create all walls for a room.
      * from_corner will create all walls that meet in a corner.

    :param room_pos: The position of the room in which this wall is.
    :type room_pos: (int, int)

    :param int wall: The wall index.
    """
    def __init__(self, room_pos, wall):
        self.room_pos = room_pos
        self.wall = wall

    __slots__ = (
        'wall',
        'room_pos')

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if self.wall == other.wall and self.room_pos == other.room_pos:
            return True
        if self.wall == other._get_back_index() \
                and all(s + d == o for s, d, o in zip(
                    self.room_pos,
                    self.direction,
                    other.room_pos)):
            return True

    def __int__(self):
        return self.wall

    def __str__(self):
        return self.NAMES[self.wall] + '@' + str(self.room_pos)
    __repr__ = __str__

    @classmethod
    def from_direction(self, room_pos, direction):
        """Creates a new wall from a direction.

        :param room_pos: The position of the room.
        :type room_pos: (int, int)

        :param int direction: The direction of the wall.

        :return: a new wall
        :rtype: self

        :raises ValueError: if the direction is invalid
        """
        return self(room_pos, self._DIRECTIONS.index(direction))

    @classmethod
    def from_room_pos(self, room_pos):
        """Generates all walls of a room.

        :param room_pos: The room coordinates.
        :type room_pos: (int, int)
        """
        for wall in self.WALLS:
            yield self(room_pos, wall)

    @classmethod
    def from_corner(self, room_pos, wall_index):
        """Generates all walls that meet in the corner where the wall has its
        start span.

        The walls are generated counter-clockwise, starting with the wall
        described by the parameters.

        :param room_pos: The position of the room.
        :type room_pos: (int, int)

        :param int wall_index: The index of the wall.
        """
        wall = start_wall = self(room_pos, wall_index)

        while True:
            yield wall

            back = wall.back
            next = self(back.room_pos, (back.wall + 1) % len(self.WALLS))
            if next == start_wall:
                break

            wall = next

    def _get_opposite_index(self):
        """Returns the index of the opposite wall.

        The opposite wall is the wall in the same room with a span opposing
        this wall.

        :return: the opposite wall
        :rtype: BaseWall
        """
        return (self.wall + len(self.WALLS) // 2) % len(self.WALLS)

    def _get_back_index(self):
        """Returns the index of the wall on the back of this wall.

        :return: the back wall
        :rtype: BaseWall
        """
        return self._get_opposite_index()

    def _get_opposite(self):
        """Returns the opposite wall.

        The opposite wall is the wall in the same room with a span opposing
        this wall.

        :return: the opposite wall
        :rtype: BaseWall

        :raise ValueError: if no opposite room exists
        """
        return self.__class__(self.room_pos, self._get_opposite_index())

    def _get_direction(self):
        """Returns a direction vector to move in when going through the wall.

        :return: a direction vector though the wall
        :rtype: (int, int)
        """
        return self._DIRECTIONS[self.wall]

    def _get_span(self):
        """Returns the span of the wall, expressed as degrees.

        The start of the wall is defined as the most counter-clockwise edge
        of the wall and the end as the start of the next wall.

        The origin of the coordinate system is the centre of the room; thus
        points to the left of the centre of room will have negative
        x-coordinates and points below the center of the room negative
        y-coordinates.

        :return: the span expressed as (start_angle, end_angle)
        :rtype: (float, float)
        """
        start = self._ANGLES[self.wall]
        end = self._ANGLES[(self.wall + 1) % len(self._ANGLES)]

        return (start, end)

    @property
    def opposite(self):
        """The opposite wall in the same room."""
        return self._get_opposite()

    @property
    def direction(self):
        """The direction vector to move in when going through the wall."""
        return self._get_direction()

    @property
    def back(self):
        """The wall on the other side of the wall."""
        direction = self._get_direction()
        return self.__class__(
            (
                self.room_pos[0] + direction[0],
                self.room_pos[1] + direction[1]),
            self._get_back_index())

    @property
    def corner_walls(self):
        """All walls in the corner that contains the start span of this
        wall."""
        return self.__class__.from_corner(self.room_pos, self.wall)

    @property
    def span(self):
        """The span of this wall"""
        return self._get_span()


class Room(object):
    """A room is a part of the maze.

    A room has a set of walls. Walls in the set are considered to have doors.

    In addition to the methods defined, the following constructs are allowed:

    * if room: => if bool(room.doors):
    * if wall in room: => if wall.has_door(wall):
    * if room[Wall.LEFT]: => if room.has_door(wall):
    * room[Wall.LEFT] = bool => room.set_door(Wall.LEFT, bool)
    * room += Wall.LEFT => room.add_door(Wall.LEFT)
    * room -= Wall.LEFT => room_remove_door(Wall.LEFT)
    """
    def __init__(self):
        self.doors = set()

    def __bool__(self):
        return bool(self.doors)
    __nonzero__ = __bool__

    def __eq__(self, other):
        return other.doors == self.doors

    def __contains__(self, wall_index):
        return self.has_door(int(wall_index))

    def __getitem__(self, wall_index):
        return self.has_door(int(wall_index))

    def __setitem__(self, wall_index, has_door):
        self.set_door(int(wall_index), has_door)

    def __iadd__(self, wall_index):
        self.add_door(int(wall_index))
        return self

    def __isub__(self, wall_index):
        self.remove_door(int(wall_index))
        return self

    def has_door(self, wall_index):
        """Returns whether a wall has a door.

        :param int wall_index: The wall to check.

        :return: whether the wall has a door
        :rtype: bool

        :raises IndexError: if wall is not a valid wall
        """
        return int(wall_index) in self.doors

    def add_door(self, wall_index):
        """Adds a door.

        :param int wall_index: The wall to which to add a door.

        :raises IndexError: if wall_index is not a valid wall
        """
        self.doors.add(int(wall_index))

    def remove_door(self, wall_index):
        """Removes a door.

        :param int wall_index: The wall from which to remove a door.

        :raises IndexError: if wall_index is not a valid wall
        """
        self.doors.discard(int(wall_index))

    def set_door(self, wall_index, has_door):
        """Adds or removes a door depending on has_door.

        :param int wall_index: The wall to modify.

        :param bool has_door: Whether to add or remove the door.

        :raises IndexError: if wall_index is not a valid wall
        """
        if has_door:
            self.add_door(int(wall_index))
        else:
            self.remove_door(int(wall_index))


class BaseMaze(object):
    """A maze is a grid of rooms.

    In addition to the methods defined, the following constructs are allowed:

    * maze[room_pos] => maze.rooms[room_pos[1]][room_pos[0]]
    * if room_pos in maze: => if room_pos[0] >= 0 and room_pos[1] >= 0
      and room_pos[0] < maze.width and room_pos[1] < maze.height
    * maze[room_pos1:room_pos2] = True => maze.add_door(room_pos1, room_pos2)
    * maze[room_pos1:room_pos2] = True => maze.remove_door(room_pos1, room_pos2)
    * maze[room_pos1:room_pos2] => maze.walk_path(room_pos1, room_pos2)
    * for room_pos in maze: => for room_pos in (rp for rp in maze.room_positions
      if maze[rp]):

    :param int width: The width of the maze.

    :param int height: The height of the maze.
    """
    def __init__(self, width, height):
        self.rooms = [[self.__class__.Room() for x in range(width)]
            for y in range(height)]
        self.width = width
        self.height = height

    Room = Room
    Wall = BaseWall

    def __getitem__(self, room_pos):
        if isinstance(room_pos, tuple) and len(room_pos) == 2:
            # A request for a specific room
            room_x, room_y = room_pos
            return self.rooms[room_y][room_x]

        if isinstance(room_pos, slice):
            # A request for the path between two rooms
            if not room_pos.step is None:
                raise ValueError()
            from_pos, to_pos = room_pos.start, room_pos.stop
            return self.walk_path(from_pos, to_pos)

        raise TypeError()

    def __setitem__(self, room_pos, value):
        if isinstance(room_pos, slice):
            # A request to set the wall between two rooms
            from_pos, to_pos = room_pos.start, room_pos.stop
            if value:
                self.add_door(from_pos, to_pos)
            else:
                self.remove_door(from_pos, to_pos)
            return

        raise TypeError()

    def __contains__(self, item):
        if isinstance(item, tuple) and len(item) == 2:
            x, y = item
            return x >= 0 and x < self.width and y >= 0 and y < self.height

        if isinstance(item, self.Wall):
            x, y = item.room_pos
            return x >= 0 and x < self.width and y >= 0 and y < self.height

    def __iter__(self):
        return (room_pos for room_pos in self.room_positions if self[room_pos])

    @property
    def room_positions(self):
        """A generator that yields the positions of all rooms"""
        for x in range(0, self.width):
            for y in range(0, self.height):
                yield (x, y)

    @property
    def edge_walls(self):
        """A generator that yields all walls on the edge of the maze; the order
        is undefined"""
        for y in range(0, self.height):
            row = (0,) if self.width == 1 \
                else (0, self.width - 1) if y == 0 or y == self.height - 1 \
                else range(0, self.width)
            for x in row:
                for wall in self.walls((x, y)):
                    if self.edge(wall):
                        yield wall

    def add_door(self, from_pos, to_pos):
        """Adds a door between two rooms.

        :param from_pos: The coordinate of the first room.
        :type from_pos: (int, int)

        :param to_pos: The coordinate of the second room.
        :type to_pos: (int, int)

        :raises IndexError: if a room lies outside of the maze

        :raises ValueError: if the rooms are not adjacent
        """
        direction = (to_pos[0] - from_pos[0], to_pos[1] - from_pos[1])
        wall = self.__class__.Wall.from_direction(from_pos, direction)

        self.set_door(from_pos, wall, True)

    def remove_door(self, from_pos, to_pos):
        """Removes a door between two rooms.

        :param from_pos: The coordinate of the first room.
        :type from_pos: (int, int)

        :param to_pos: The coordinate of the second room.
        :type to_pos: (int, int)

        :raises IndexError: if a room lies outside of the maze

        :raises ValueError: if the rooms are not adjacent
        """
        direction = (to_pos[0] - from_pos[0], to_pos[1] - from_pos[1])
        wall = self.__class__.Wall.from_direction(from_pos, direction)

        return self.set_door(from_pos, wall, False)

    def set_door(self, room_pos, wall, has_door):
        """Adds or removes a door.

        :param room_pos: The coordinates of the room.
        :type room_pos: (int, int)

        :param BaseWall wall: The wall to modify.

        :param bool has_door: True to add the door and False to remove it.

        :raises IndexError: if room_pos lies outside of the maze
        """
        if not room_pos in self:
            raise IndexError()

        # Get the coordinate of the other room
        if not isinstance(wall, self.Wall):
            wall = self.Wall(room_pos, int(wall))
        other_wall = wall.back
        to_pos = other_wall.room_pos

        self[room_pos][wall] = has_door

        if to_pos in self:
            self[to_pos][other_wall] = has_door

    def get_center(self, room_pos):
        """Returns the physical coordinates of the centre of a room.

        If lines are drawn on a circle with radius 1.0 centred on this point
        between the angles corresponding to the spans of the walls, the lines of
        adjacent rooms will cover each other.

        :param room_pos: The position of the room.
        :type room_pos: (int, int)

        :return: the coordinates of the room
        :rtype: (float, float)

        :raises IndexError: if a room lies outside of the maze
        """
        raise NotImplementedError()

    def adjacent(self, room1_pos, room2_pos):
        """Returns whether two rooms are adjacent.

        The rooms are adjacent if the room at room1_pos has a wall that leads to
        the room at room2_pos.

        :param room1_pos: The coordinate of the first room.
        :type room1_pos: (int, int)

        :param room2_pos: The coordinate of the second room.
        :type room2_pos: (int, int)

        :return: whether there is a wall between the two rooms
        :rtype: bool
        """
        return any((room1_pos[0] + wall.direction[0],
                room1_pos[1] + wall.direction[1]) == room2_pos
            for wall in self.walls(room1_pos))

    def connected(self, room1_pos, room2_pos):
        """Returns whether two rooms are connected by a single wall containing a
        door.

        :param room1_pos: The coordinate of the first room.
        :type room1_pos: (int, int)

        :param room2_pos: The coordinate of the second room.
        :type room2_pos: (int, int)

        :return: whether the two rooms are adjacent and have doors
        :rtype: bool
        """
        # Make sure that they are adjacent
        if not self.adjacent(room1_pos, room2_pos):
            return False

        # Make sure the wall has a door
        try:
            direction = (
                room2_pos[0] - room1_pos[0],
                room2_pos[1] - room1_pos[1])
            wall_index = int(self.__class__.Wall.from_direction(
                room1_pos, direction))
            return wall_index in self[room1_pos]
        except ValueError:
            return False

    def edge(self, wall):
        """Returns whether a wall is on the edge of the maze.

        :param BaseWall wall: The wall.

        :return: whether the wall is on the edge of the maze
        :rtype: bool
        """
        return wall in self and not wall.back in self

    def walls(self, room_pos):
        """Generates all walls of a room.

        :param room_pos: The coordinates of the room.
        :type room_pos: (int, int)

        :raises IndexError: if a room lies outside of the maze
        """
        if room_pos in self:
            return self.__class__.Wall.from_room_pos(room_pos)
        else:
            raise IndexError("Room %s is not part of the maze" % str(room_pos))

    def doors(self, room_pos):
        """Generates all walls with doors.

        :param room_pos: The coordinates of the room.
        :type room_pos: (int, int)

        :raises IndexError: if a room lies outside of the maze
        """
        room = self[room_pos]

        for wall in self.__class__.Wall.WALLS:
            if wall in room:
                yield self.__class__.Wall(room_pos, wall)

    def walk_from(self, room_pos, wall, require_door = False):
        """Returns the coordinates of the room through the specified wall.

        The starting room, room_pos, may be outside of the maze if it is
        immediately on the edge and the movement is inside the maze.

        :param room_pos: The room from which to walk.
        :type room_pos: (int, int)

        :param wall: The wall to walk through.
        :type wall: BaseWall or int

        :param bool require_door: Whether to require a door in the specified
            direction.

        :return: the destination coordinates
        :rtype: (int, int)

        :raises ValueError: if require_door is True and there is no door on the
            wall

        :raises IndexError: if the destination room lies outside of the maze
        """
        wall = self.__class__.Wall(room_pos, int(wall))
        direction = wall.direction
        result = (room_pos[0] + direction[0], room_pos[1] + direction[1])

        if require_door:
            if not wall in self[room_pos]:
                raise ValueError('(%d, %d) is not inside the maze' % room_pos)

        if result in self:
            return result
        else:
            raise IndexError()

    def walk(self, wall, require_door = False):
        """Returns the coordinates of the room through the specified wall.

        :param wall: The wall to walk through.
        :type wall: BaseWall

        :param bool require_door: Whether to require a door in the specified
            direction.

        :return: the destination coordinates
        :rtype: (int, int)

        :raises ValueError: if require_door is True and there is no door on the
            wall

        :raises IndexError: if the destination room lies outside of the maze
        """
        return self.walk_from(wall.room_pos, int(wall), require_door)

    def walk_path(self, from_pos, to_pos, visitor = lambda room_pos: None):
        """Generates all rooms on the shortest path between two rooms.

        :param from_pos: The room in which to start. This room is included.
        :type room_pos: (int, int)

        :param to_pos: The last room to visit.
        :type to_pos: (int, int)

        :param visitor: A callback to call for every room encountered. This
            callback may raise StopIteration to cancel the walking.
        :type visitor: func((int, int))

        :raises ValueError: if there is no path between the rooms
        """
        # Handle walking to the same room efficiently
        if from_pos == to_pos:
            visitor(from_pos)
            yield from_pos
            return

        # Swap from_pos and to pos to make reconstructing the path easier
        from_pos, to_pos = to_pos, from_pos

        def h(room_pos):
            """The heuristic for a room"""
            return sum((t - f)**2 for f, t in zip(room_pos, to_pos))

        # The rooms already evaluated
        closed_set = []

        # The rooms pending evaluation; this list is sorted on cost
        class infinity(object):
            def __cmp__(self, other):
                if isinstance(other, type(self)): return 0
                else: return 1
        open_set = [(infinity(), from_pos)]

        # The cost from from_pos to the room along the best known path
        g_score = {}
        g_score[from_pos] = 0

        # The estimated cost from from_pos to to_pos through the room
        f_score = {}
        f_score[from_pos] = h(from_pos)

        # The room from which we entered a room
        came_from = {}

        while open_set:
            # Get the node in open_set having the lowest f_score value
            cost, current = open_set.pop(0)

            # Visit the room first
            visitor(current)

            # Have we reached the goal?
            if current == to_pos:
                while current != from_pos:
                    yield current
                    current = came_from[current]
                yield from_pos
                return

            closed_set.append(current)
            for wall in self.doors(current):
                next = wall.back.room_pos

                # Ignore rooms already evaluated or rooms outside of the maze
                if next in closed_set or not next in self:
                    continue

                # The cost to get to this room is one more that the room from
                # which we came
                g = g_score[current] + 1

                # Is this a new room, or has the score improved since last?
                if not next in open_set or g < g_score[next]:
                    came_from[next] = current
                    g_score[next] = g
                    f = g + h(next)
                    f_score[next] = f

                    # Insert the next room while keeping open_set sorted
                    if not next in open_set:
                        bisect.insort(open_set, (f, next))

        raise ValueError()




# coding=utf-8
'''
pymaze
Copyright (C) 2012-2014 Moses Palmér

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
'''

def initialize(maze, randomizer):
    """
    A function that initialises a maze with the randomised prim algorithm.

    See http://en.wikipedia.org/wiki/Maze_generation_algorithm

    @param maze
        The maze to initialise.
    @param randomizer
        The function used as a source of randomness. It will be called with an
        argument describing the maximum value to return. It may return any
        integers between 0 and the non-inclusive maximum value.
    """
    # Start with a random room and add all its walls except those on the edge
    start_x, start_y = randomizer(maze.width), randomizer(maze.height)
    walls = [wall for wall in maze.walls((start_x, start_y))
        if not maze.edge(wall)]

    while walls:
        # Select a random wall
        index = randomizer(len(walls))
        wall = walls.pop(index)

        # Get the room behind the wall
        next_room_pos = maze.walk(wall)

        # Is this the first time we visit this room?
        if not maze[next_room_pos]:
            # Add a door to the wall
            maze.set_door(wall.room_pos, wall, True)

            # Add all walls of the new room except those leading to rooms
            # already visited or leading out of the maze
            for w in maze.walls(next_room_pos):
                try:
                    if not maze[maze.walk(w)]:
                        walls.append(w)
                except IndexError:
                    pass




# coding=utf-8
'''
pymaze
Copyright (C) 2012-2014 Moses Palmér

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
'''

try:
    import cairocffi as cairo
except ImportError:
    import cairo
import math
import sys

def calculate_bounds(maze):
    """
    Calculates the bounds of the walls of a maze.

    @param maze
        The maze whose bounds to calculate.
    @return the tuple (min_x, min_y, max_x, max_y)
    """
    class infinity(object):
        def __cmp__(self, other):
            if isinstance(other, type(self)): return 0
            else: return 1
    max_x, max_y = 0, 0
    min_x, min_y = infinity(), infinity()
    for wall in maze.edge_walls:
        a = wall.span[0]
        cx, cy = maze.get_center(wall.room_pos)
        px, py = cx + math.cos(a), cy + math.sin(a)
        max_x, max_y = max(max_x, px), max(max_y, py)
        min_x, min_y = min(min_x, px), min(min_y, py)
    return (min_x, min_y, max_x, max_y)

def draw_walls(maze, ctx, coords):
    """
    Draws the walls of a maze.

    @param maze
        The maze whose walls to draw.
    @param ctx
        The cairo context to use for drawing.
    @param coords
        A callable that transforms its parameters (maze_x, maze_y) to
        coordinates in the cairo context.
    """
    # Note that we have not yet painted any walls for any rooms
    for room_pos in maze.room_positions:
        maze[room_pos].painted = set()

    # Initialise the wall queue
    queue = []
    def extend_queue():
        """
        Finds a room with walls that have not yet been painted and adds them to
        queue.
        """
        for room_pos in maze.room_positions:
            remaining = [w
                for w in maze.walls(room_pos)
                if not int(w) in maze[w.room_pos]
                    and not int(w) in maze[w.room_pos].painted]
            if remaining:
                queue.extend(remaining)
                break
    extend_queue()

    # Draw the walls
    needs_move = True
    while queue:
        # Get the last wall from the queue and retrieve all remaining walls in
        # its corner; we add the back of the walls in order to actually move
        # along the wall instead of just spinning around the corner
        wall = queue.pop()
        walls = [w.back if w.back in maze else w
            for w in wall.corner_walls
            if w in maze or w.back in maze]
        remaining = [w for w in walls
            if not int(w) in maze[w.room_pos]
                and not int(w) in maze[w.room_pos].painted
                and not w == wall]

        # Queue all remaining walls for later use
        queue.extend(remaining)

        start_angle, end_angle = wall.span
        offset_x, offset_y = maze.get_center(wall.room_pos)

        def angle_to_coordinate(angle):
            """
            Converts an angle from a wall span in the current room to a
            coordinate.

            @param angle
                The angle to convert.
            @return the tuple (x, y)
            """
            return coords(
                offset_x + math.cos(angle),
                offset_y + math.sin(angle))

        # If we need to move, we move to the end of the span since
        # maze.Wall.from_corner will yield walls with the start span in the
        # given corner
        if needs_move:
            ctx.move_to(*angle_to_coordinate(end_angle))
        ctx.line_to(*angle_to_coordinate(start_angle))

        # Mark the current wall as painted, and the wall on the other side as
        # well as long as this is not a wall along the edge of the maze
        maze[wall.room_pos].painted.add(int(wall))
        if not maze.edge(wall):
            maze[wall.back.room_pos].painted.add(int(wall.back))

        # If we have reached a dead end, we need to stroke the line and start
        # over with a wall from the queue
        if not remaining:
            ctx.stroke()
            needs_move = True
        else:
            needs_move = False

        # If the queue is empty, check if any walls remain
        if not queue:
            extend_queue()

    ctx.stroke()

    # Remove the temporary attribute
    for room_pos in maze.room_positions:
        del maze[room_pos].painted

def draw_path_smooth(maze, ctx, coords, solution):
    """
    Draws the solution path using a smooth bezier curve.

    @param maze
        The maze whose solution to draw.
    @param ctx
        The cairo context to use for drawing.
    @param coords
        A callable that transforms its parameters (maze_x, maze_y) to
        coordinates in the cairo context.
    @param solution
        The solution. This must be a list of all rooms to traverse.
    """
    room_positions = ((solution[i - 1], solution[i], solution[i + 1])
        for i in range(1, len(solution) - 1))
    ctx.move_to(*coords(*maze.get_center(solution[0])))
    for (previous_pos, current_pos, next_pos) in room_positions:
        # Draw a bezier curve from the wall to the previous room to the wall
        # to the next room
        previous_center = coords(*maze.get_center(previous_pos))
        current_center = coords(*maze.get_center(current_pos))
        next_center = coords(*maze.get_center(next_pos))
        d = 0.3
        ctx.curve_to(
            d * previous_center[0] + (1.0 - d) * current_center[0],
            d * previous_center[1] + (1.0 - d) * current_center[1],
            d * next_center[0] + (1.0 - d) * current_center[0],
            d * next_center[1] + (1.0 - d) * current_center[1],
            0.5 * (current_center[0] + next_center[0]),
            0.5 * (current_center[1] + next_center[1]))
    ctx.line_to(*coords(*maze.get_center(solution[-1])))
    ctx.stroke()

def draw_path(maze, ctx, coords, solution):
    """
    Draws the solution path using straight lines.

    @param maze
        The maze whose solution to draw.
    @param ctx
        The cairo context to use for drawing.
    @param coords
        A callable that transforms its parameters (maze_x, maze_y) to
        coordinates in the cairo context.
    @param solution
        The solution. This must be a list of all rooms to traverse.
    """
    for i, room_pos in enumerate(solution):
        # Draw a line to the centre of the room
        center = coords(*maze.get_center(room_pos))
        if i == 0:
            ctx.move_to(*center)
        else:
            ctx.line_to(*center)
    ctx.stroke()

def make_image(maze, solution, room_size, output, background_color, wall_color,
        path_color, wall_width, path_width, path_smooth):
    room_width, room_height = room_size
    image_create, image_write = output

    # Calculate the actual size of the image
    min_x, min_y, max_x, max_y = calculate_bounds(maze)

    # Create the cairo surface and context
    width = int((max_x - min_x) * room_width) + 2 * wall_width + 1
    height = int((max_y - min_y) * room_height) + 2 * wall_width + 1
    surface = image_create(width, height)
    ctx = cairo.Context(surface)

    # Make line caps and line joins round
    ctx.set_line_cap(cairo.LINE_CAP_ROUND)
    ctx.set_line_join(cairo.LINE_JOIN_ROUND)

    # Clear the background
    ctx.set_source_rgba(*background_color)
    ctx.paint()

    def coords(x, y):
        """
        Converts coordinates from maze coordinates to context coordinates.

        The coordinates are scaled according to the room dimesions and offset
        with the minimum coordinates in the respective dimension and then
        rounded to make the lines sharp when drawing.

        @param x, y
            The maze coordinates to convert.
        @return coordinates to use for drawing
        """
        return (
            wall_width + (
                round((x - min_x) * room_width)),
            height - wall_width - (
                round((y - min_y) * room_height)))

    # Draw the walls
    ctx.set_source_rgba(*wall_color)
    ctx.set_line_width(wall_width)
    draw_walls(maze, ctx, coords)

    # Draw the path
    ctx.set_source_rgba(*path_color)
    ctx.set_line_width(path_width)
    if path_smooth:
        draw_path_smooth(maze, ctx, coords, solution)
    else:
        draw_path(maze, ctx, coords, solution)

    image_write(surface)




#!/usr/bin/env python
# coding: utf8

import os
import setuptools
import sys


# The name of the package on PyPi
PYPI_PACKAGE_NAME = 'pymaze'

# The name of the main Python package
MAIN_PACKAGE_NAME = 'maze'

# The package URL
PACKAGE_URL = 'https://github.com/moses-palmer/maze.py'

SCRIPTS = ['tools/amaze']

# The author email
AUTHOR_EMAIL = 'moses.palmer@gmail.com'


def setup(**kwargs):
    setuptools.setup(
        name = PYPI_PACKAGE_NAME,
        version = '.'.join(str(i) for i in INFO['version']),
        description = ''
            'A library to generate and display mazes',
        long_description = README + '\n\n' + CHANGES,

        install_requires = [
            'cairocffi >=0.6'],
        setup_requires = [],

        author = INFO['author'],
        author_email = AUTHOR_EMAIL,

        url = PACKAGE_URL,

        packages = setuptools.find_packages(
            os.path.join(
                os.path.dirname(__file__),
                'lib'),
            exclude = ['tests', 'tests.suites']),
        scripts = SCRIPTS,
        package_dir = {'': 'lib'},
        zip_safe = True,

        license = 'GPLv3',
        classifiers = [],

        **kwargs)


# Arguments passed to setup
setup_arguments = {}


# Read globals from ._info without loading it
INFO = {}
with open(os.path.join(
        os.path.dirname(__file__),
        'lib',
        'maze',
        '_info.py')) as f:
    for line in f:
        try:
            name, value = (i.strip() for i in line.split('='))
            if name.startswith('__') and name.endswith('__'):
                INFO[name[2:-2]] = eval(value)
        except ValueError:
            pass


try:
    # Read README
    with open(os.path.join(
            os.path.dirname(__file__),
            'README.rst')) as f:
        README = f.read()


    # Read CHANGES
    with open(os.path.join(
            os.path.dirname(__file__),
            'CHANGES.rst')) as f:
        CHANGES = f.read()
except IOError:
    README = ''
    CHANGES = ''


try:
    setup(**setup_arguments)
except Exception as e:
    try:
        sys.stderr.write(e.args[0] % e.args[1:] + '\n')
    except:
        sys.stderr.write(str(e) + '\n')



# coding=utf-8
'''
pymaze
Copyright (C) 2012-2014 Moses Palmér

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
'''

import math

from . import BaseWall, BaseMaze

class QuadWall(BaseWall):
    # The angles for each corner
    _ANGLES = []

    # The directions for each wall
    _DIRECTIONS = []

    # The names of the walls
    NAMES = []

    # The list of walls; this is the list [0, 1, 2... N]
    WALLS = []

    # The scale factor when converting maze coordinates to physical coordinates
    MULTIPLICATOR = 2.0 / math.sqrt(2.0)

    __slots__ = tuple()

    start_angle = (5 * math.pi) / 4
    data = (
        ('LEFT', (-1, 0)),
        ('UP', (0, 1)),
        ('RIGHT', (1, 0)),
        ('DOWN', (0, -1)))
    for i, (name, dir1) in enumerate(data):
        locals()[name] = i

        next_angle = _ANGLES[-1] - 2 * math.pi / len(data) \
            if _ANGLES else start_angle
        while next_angle < 0.0:
            next_angle += 2 * math.pi
        _ANGLES.append(next_angle)

        _DIRECTIONS.append(dir1)

        NAMES.append(name.lower())
        WALLS.append(i)

class Maze(BaseMaze):
    """A maze with square rooms.

    This is the traditional maze. Maze coordinates correspond to physical
    coordinates after a simple scale operation.
    """
    Wall = QuadWall

    def get_center(self, room_pos):
        return (
            (room_pos[0] + 0.5) * self.Wall.MULTIPLICATOR,
            (room_pos[1] + 0.5) * self.Wall.MULTIPLICATOR)
